package itemhandler;

import static staticconstants.InvalidReturn.INVALID_CATEGORY_MESSAGE;
import static staticconstants.InvalidReturn.INVALID_COLOR_MESSAGE;
import static staticconstants.InvalidReturn.INVALID_DOUBLE_MESSAGE;
import static staticconstants.InvalidReturn.INVALID_INT_MESSAGE;
import static staticconstants.InvalidReturn.INVALID_POSITIVE_OR_NEGATIVE_INT_MESSAGE;
import static staticconstants.InvalidReturn.INVALID_STRING_MESSAGE;
import static staticconstants.InvalidReturn.ITEM_NUMBER_OCCUPIED_IN_REGISTRY;
import static staticconstants.InvalidReturn.NUMBER_EXCEPTION_ERROR_MESSAGE;

import java.security.InvalidParameterException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Scanner;
import staticconstants.Category;
import staticconstants.ItemFieldEnum;
import staticconstants.ValidColors;
import utility.ItemFilter;
import utility.ItemSorter;
import utility.Parameter;

/**
 * User interface class handles the usage of all other classes except SmartHouseApp and allows
 * user to change and see item registry filtered, sorted or plain based on user input from terminal
 * through Scanner.
 */
public class UserInterface
{
    public final Scanner scannerInput = new Scanner(System.in);
    private Map<String, Item> itemRegistry = new HashMap<>();

    /** method used to retrieve registry.
     *
     */
    private void retrieveItemRegistry()
    {
        try
        {
            ItemRegistryStorage saveHashMap = new ItemRegistryStorage();
            itemRegistry = saveHashMap.readFile();
        } catch (Exception e)
        {
            System.err.println("no Items were found in registry: " + e);
        }
    }

    /**
     * Method that gets user input in a loop that ends after user has typed in a valid
     * String.

     * @param userMessage The String that is printed to tell the user what parameter is expected
     * @param isOptional  boolean that is true for an optional parameter and false for a mandatory
     *                    parameter
     * @return input the valid user input that is to be used and handled to construct a new item
     *      or change an already existing one through an augmenter.
     */
    private String getUserInputString(String userMessage, boolean isOptional)
    {
        String input;
        boolean flag;
        do
        {
            System.out.println(userMessage);
            input = scannerInput.nextLine();
            if (isOptional && input.equals(""))
            {
                input = null;
                flag = true;
            } else if (!Parameter.checkValid(input))
            {
                System.err.println(INVALID_STRING_MESSAGE);
                flag = false;
            } else
            {
                flag = true;
            }
        } while (!flag);
        return input;
    }

    private String getUserInputItemNumber()
    {
        String userMessage = "Type item number, item number must be unique to the item registry"
            + " (this parameter cannot be the same as another items number)";
        String input;
        boolean flag;
        do
        {
            System.out.println(userMessage);
            input = scannerInput.nextLine();

            if (!Parameter.checkValid(input))
            {
                System.err.println(INVALID_STRING_MESSAGE);
                flag = false;
            } else if (itemRegistry.get(input) != null)
            {
                System.err.println(ITEM_NUMBER_OCCUPIED_IN_REGISTRY);
                flag = false;
            } else
            {
                flag = true;
            }
        } while (!flag);
        return input;
    }

    /**
     * Method that gets user input in a loop that ends after user has typed in a valid
     * color.
     *
     * @return input valid color
     */
    private String getUserInputColor(boolean isOptional)
    {
        String input;
        String userMessage = "Type item Color valid colors are: "
            + Arrays.stream(ValidColors.values()).toList();
        boolean flag;
        do
        {
            System.out.println(userMessage);
            input = scannerInput.nextLine();

            if (isOptional && input.equals(""))
            {
                input = null;
                flag = true;
            } else if (!Parameter.checkValid(input) || !Parameter.checkValidColor(input))
            {
                System.err.println(INVALID_COLOR_MESSAGE);
                flag = false;
            } else
            {
                flag = true;
            }
        } while (!flag);
        return input;
    }

    /**
     * Method that gets user input in a loop that ends after user has typed in a valid
     * int.

     * @param userMessage The String that is printed to tell the user what parameter is expected
     * @return input valid number
     */
    private int getUserInputInt(String userMessage)
    {
        int input;
        do
        {
            try
            {
                System.out.println(userMessage);
                input = Integer.parseInt(scannerInput.nextLine());

                if (!Parameter.checkValid(input))
                {
                    System.err.println(INVALID_INT_MESSAGE);
                }

            } catch (Exception e)
            {
                System.err.println(NUMBER_EXCEPTION_ERROR_MESSAGE);
                System.err.println("" + e);
                input = 0;
            }
        } while (!Parameter.checkValid(input));
        return input;
    }

    /**
     * Method that gets user input in a loop that ends after user has typed in a valid
     * category.
     *
     * @return input valid category
     */
    private int getUserInputItemCategory()
    {
        boolean flag;
        int input;
        String userMessage = "Type item category, valid categories are: "
            + Category.getCategoryMap();
        do
        {
            input = getUserInputInt(userMessage);
            flag = Parameter.checkValidCategory(input);
            if (!flag)
            {
                System.err.println(INVALID_CATEGORY_MESSAGE);
            }
        } while (!flag);
        return input;
    }

    /**
     * Method that gets user input in a loop that ends after user has typed in a valid
     * double used to specify the size of the item through height, length and weight.

     * @param userMessage The String that is printed to tell the user what parameter is expected
     * @return input valid category
     */
    private double getUserInputDouble(String userMessage)
    {
        double input;
        do
        {
            try
            {
                System.out.println(userMessage);
                input = Double.parseDouble(scannerInput.nextLine());

                if (!Parameter.checkValid(input))
                {
                    System.err.println(INVALID_DOUBLE_MESSAGE);
                }

            } catch (Exception e)
            {
                System.err.println(NUMBER_EXCEPTION_ERROR_MESSAGE);
                System.err.println("" + e);
                input = 0.0;
            }
        } while (!Parameter.checkValid(input));
        return input;
    }

    /**
     * Uses Scanner to add parameters from input through the getUserInput methods.
     * Uses Item class constructor to create a new item.
     *
     * @return new Item
     */
    private Item createItemFromInput()
    {
        String itemNumber = getUserInputItemNumber();
        String itemColor = getUserInputColor(true);
        String itemDescription = getUserInputString("Type item description (optional)",
            true);
        String itemBrand = getUserInputString("Type item brand (optional)", true);

        int itemPrice = getUserInputInt("Type item price");
        int itemCategory = getUserInputItemCategory();
        int itemQuantity =
            getUserInputInt("Type number of items currently in stock");

        double itemWeight = getUserInputDouble("Type item weight (can be a decimal)");
        double itemHeight = getUserInputDouble("Type item height (can be a decimal)");
        double itemLength = getUserInputDouble("Type item length (can be a decimal)");
        Item item = new Item(itemNumber, itemPrice,
            itemCategory, itemQuantity, itemWeight, itemLength, itemHeight);
        if (itemColor != null)
        {
            item.setColor(itemColor);
        }
        if (itemDescription != null)
        {
            item.setDescription(itemDescription);
        }
        if (itemBrand != null)
        {
            item.setBrand(itemBrand);
        }
        System.out.println("Item has been created!");
        return item;
    }

    private void addOrSubtractQuantity(Item itemToAddOrSubtract)
    {
        boolean isValidInput = false;
        int currentItemQuantity = itemToAddOrSubtract.getQuantity();
        int numberToAddOrSubtractWith;
        int possibleNewItemQuantity;
        do
        {
            System.out.println("Type the amount you would like to change the "
                + "number of items in stock/quantity with\n "
                + "(negative number to subtract, positive number to add)\n"
                + "current number of items in stock is: " + itemToAddOrSubtract.getQuantity());
            try
            {
                numberToAddOrSubtractWith = Integer.parseInt(scannerInput.nextLine());
                possibleNewItemQuantity = currentItemQuantity
                    + numberToAddOrSubtractWith;
                if (possibleNewItemQuantity >= 0)
                {
                    itemToAddOrSubtract.setQuantity(currentItemQuantity
                        + numberToAddOrSubtractWith);
                    isValidInput = true;
                } else
                {
                    System.err.println("The number you intend to subtract number of items in stock"
                        + " with is larger than the current number of items in stock");
                }
            } catch (Exception e)
            {
                System.err.println(INVALID_POSITIVE_OR_NEGATIVE_INT_MESSAGE);
            }
        } while (!isValidInput);
    }

    /**
     * Updates the item data.
     *
     * @param item imported itemPackage.Item object to be changed
     */
    private void changeItemData(Item item)
    {
        String typedParameterName;
        do
        {
            System.out.println(
                "type enter to cancel, type the name of the data you wish to change.");
            System.out.println("available parameters to change are: "
                + Arrays.stream(ItemFieldEnum.values()).toList());
            typedParameterName = scannerInput.nextLine().toUpperCase();
            try
            {
                switch (ItemFieldEnum.valueOf(typedParameterName))
                {
                    case QUANTITY -> addOrSubtractQuantity(item);
                    case PRICE -> item.setPrice(getUserInputInt("type item price"));
                    case CATEGORY -> item.setCategory(getUserInputItemCategory());
                    case NUMBER -> System.err.println(
                        "Item number is permanent! You may need to delete this item"
                            + " then create a new one to have the same item with different item "
                            + "number");
                    case BRAND -> item.setBrand(getUserInputString("Type item brand",
                        false));
                    case DESCRIPTION -> item.setDescription(getUserInputString(
                        "Type item description", false));
                    case COLOR -> item.setColor(getUserInputColor(
                        false));
                    case WEIGHT -> item.setWeight(getUserInputDouble("Type item weight"));
                    case LENGTH -> item.setLength(getUserInputDouble("Type item length"));
                    case HEIGHT -> item.setHeight(getUserInputDouble(
                        "Type item height"));
                    default -> System.out.println(
                        "no parameter with that name were found, please retry");
                }
            } catch (IllegalArgumentException e)
            {
                if (!typedParameterName.equals(""))
                {
                    System.err.println("No parameter with that name was found, please try again");
                }
            }
        } while (!typedParameterName.equals(""));
        System.out.println(
            "Task completed! The information you wished to change has been changed.");
    }

    /** Adds items to itemRegistry according to user input.
     *
     */
    private void addItemToRegistryFromInput()
    {
        Item item;
        boolean flag;
        do
        {
            try
            {
                item = createItemFromInput();
                System.out.println("~~~~~~~~~~~~~~~");
                flag = !this.itemRegistry.containsKey(item.getNumber());
                if (!flag)
                {
                    System.err.println("the item you have typed has the same itemNumber as"
                        + "another item from the item list, please retry");
                    System.err.println(ITEM_NUMBER_OCCUPIED_IN_REGISTRY);
                } else
                {
                    this.itemRegistry.putIfAbsent(item.getNumber(), item);
                    System.out.println(
                        "Type enter to exit item creator, otherwise type anything else");
                }
            } catch (InvalidParameterException e)
            {
                System.err.println("This item has an invalid parameter, please retry");
                flag = false;
            }
        } while (!scannerInput.nextLine().equals("") || !flag);
    }

    /**
     * Selects item from registry to get or to set item parameters (it allows user to access
     * a specific item).
     */
    private void selectItemToAugment()
    {
        String userInput;
        System.out.println("Type the itemNumber of the item you wish to access. Available "
            + "itemNumbers are as follows:");
        System.out.println(itemRegistry.keySet());
        userInput = scannerInput.nextLine();
        if (itemRegistry.containsKey(userInput))
        {
            changeItemData(itemRegistry.get(userInput));
        } else
        {
            System.err.println("No item with given itemNumber were found");
        }
    }

    /**
     * Deletes items from users registry according to user input.
     *
     */
    private void deleteItemFromItemRegistry()
    {
        String userInputItemNumber;
        do
        {
            System.out.println("Type the itemNumber of the item you wish to delete");
            System.out.println(itemRegistry.keySet());
            System.out.println();
            System.out.println("Type nothing to quit item deletion");
            userInputItemNumber = scannerInput.nextLine();
            if (!itemRegistry.containsKey(userInputItemNumber))
            {
                System.err.println("No item with given itemNumber were found");
            } else
            {
                itemRegistry.remove(userInputItemNumber);
                System.out.println("The item with item number: " + userInputItemNumber
                    + " has been removed successfully!");
                System.out.println("~~~~~~~");
            }
        } while (!userInputItemNumber.equals(""));
    }

    /**
     * Uses user input to show only items containing user input in their item Description, Category
     * or Color.
     */
    private void filterInItemRegistry()
    {

        String parameterUsedToFilter;
        ItemFilter itemSorter = new ItemFilter();
        HashMap<String, Item> sortedItemsMap = new HashMap<>();
        boolean flag;
        do
        {
            flag = false;
            System.out.println("Available parameters to filter by"
                + " are as follows: Category, Brand, Color or Description");
            parameterUsedToFilter = scannerInput.nextLine();
            try
            {
                switch (ItemFieldEnum.valueOf(parameterUsedToFilter.toUpperCase()))
                {
                    case DESCRIPTION -> sortedItemsMap =
                            (HashMap<String, Item>) itemSorter.filterItemRegistry(
                                getUserInputString(
                                    "Type description to sort by", false),
                                itemRegistry, ItemFieldEnum.DESCRIPTION);
                    case COLOR -> sortedItemsMap = (HashMap<String, Item>) itemSorter
                        .filterItemRegistry(getUserInputColor(false), itemRegistry,
                            ItemFieldEnum.COLOR);
                    case CATEGORY -> sortedItemsMap =
                            (HashMap<String, Item>) itemSorter.filterItemRegistry(
                                String.valueOf(getUserInputItemCategory()), itemRegistry,
                                ItemFieldEnum.CATEGORY);
                    case BRAND -> sortedItemsMap = (HashMap<String, Item>)
                        itemSorter.filterItemRegistry(getUserInputString(
                                "Type brand to sort by", false), itemRegistry,
                            ItemFieldEnum.BRAND);
                    default ->
                    {
                        System.err.println("Neither description, brand, color or category "
                            + "were found, please retry");
                        flag = true;
                    }
                }
            } catch (IllegalArgumentException e)
            {
                System.err.println("No item parameter with given name was found, please retry");
            }
        } while (flag);
        ItemTable itemTable = new ItemTable();
        itemTable.displayItemTable(sortedItemsMap);
    }

    /**
     * Sorts and shows item registry according to Price, Height, Length, Weight or Quantity.
     */
    private void sortInItemRegistry()
    {
        String userInput;
        ItemTable itemTable = new ItemTable();
        do
        {
            System.out.println(
                "Available parameters to sort numerically by are: "
                    + "Price, Quantity, Height, Length and Weight (type enter to exit)");
            userInput = scannerInput.nextLine().toUpperCase();
            try
            {
                switch (ItemFieldEnum.valueOf(userInput))
                {
                    case PRICE -> itemTable.displayItemTable(ItemSorter.sortItemInteger(
                        itemRegistry, ItemFieldEnum.PRICE));
                    case QUANTITY -> itemTable.displayItemTable(ItemSorter.sortItemInteger(
                        itemRegistry, ItemFieldEnum.QUANTITY));
                    case HEIGHT -> itemTable.displayItemTable(ItemSorter.sortItemDouble(
                        itemRegistry, ItemFieldEnum.HEIGHT));
                    case LENGTH -> itemTable.displayItemTable(ItemSorter.sortItemDouble(
                        itemRegistry, ItemFieldEnum.LENGTH));
                    case WEIGHT -> itemTable.displayItemTable(ItemSorter.sortItemDouble(
                        itemRegistry, ItemFieldEnum.WEIGHT));
                    default -> System.err.println("You have entered a parameter that can be sorted "
                        + "by clicking on the item detail on the top of the table you can "
                        + "sort that item detail alphabetically, please retry");
                }
                System.out.println("~~~\n a new window has opened!\n~~~");
            } catch (IllegalArgumentException e)
            {
                if (!userInput.equals(""))
                {
                    System.err.println("No item parameter was found with given name, please retry");
                }
            }
        } while (!userInput.equals(""));
    }

    private void save()
    {
        System.out.println("Saving...");
        ItemRegistryStorage saveHashMap = new ItemRegistryStorage();
        try
        {
            saveHashMap.saveFile(itemRegistry);
        } catch (Exception e)
        {
            System.err.println("An error occurred during saving process");
        }
    }

    private void printItemRegistryCategory()
    {
        System.out.println("itemNumber\titemCategory");
        String itemCategory;
        for (Item item : itemRegistry.values())
        {
            itemCategory = Category.getCategoryFromNumber(item.getCategory());
            System.out.println(item.getNumber() + "\t\t\t" + itemCategory);
        }
    }

    /**
     * Sets an item with user inputted number on sale in a percentage.
     */
    private void setItemOnSale()
    {
        String userInput;
        do
        {
            System.out.println("Type the number of the item you wish to be put on sale, "
                + "type enter to exit");
            userInput = scannerInput.nextLine();
            Item item;
            try
            {
                if (itemRegistry.get(userInput) != null)
                {
                    item = itemRegistry.get(userInput);
                    System.out.println("Type the percentage of the sale by"
                        + " the current price is: " + item.getPrice());
                    float origPrice = Float.parseFloat(String.valueOf(item.getPrice()));
                    float factor = (100f - Float.parseFloat(scannerInput.nextLine())) / 100f;
                    int newPrice = Math.round(Math.round(origPrice) * factor);
                    item.setPrice(newPrice);
                    System.out.println("The price is now: " + item.getPrice());
                } else if (!userInput.equals(""))
                {
                    System.err.println("No item was found with that itemNumber");
                }
            } catch (NumberFormatException e)
            {
                System.err.println("You may have typed something other than a whole "
                    + "number please retry");
            }
        } while (!userInput.equals(""));
    }

    /**
     * Prints the different commands the user can use.
     *
     */
    private void printUserOptions()
    {
        System.out.println("The numbers you can type to execute commands are as follows:");
        System.out.println("0: Save and exit program");
        System.out.println("1: Change information from registry");
        System.out.println("2: Add more items to the item registry");
        System.out.println("3: Get all item numbers in item registry");
        System.out.println("4: Get all items' information from item registry");
        System.out.println("5: Delete item with given item number from item registry");
        System.out.println("6: Filter all the items containing given keywords in item details,"
            + " item color or item category");
        System.out.println("7: Sort registry numerically by price, quantity, weight, height"
            + "or length");
        System.out.println("8: Print all items' category");
        System.out.println("9: Put item with given number on sale");
    }

    private int userOptionHandler()
    {
        printUserOptions();
        int userInput = Integer.parseInt(scannerInput.nextLine());
        ItemTable itemTable = new ItemTable();
        switch (userInput)
        {
            case 1 -> selectItemToAugment();
            case 2 -> addItemToRegistryFromInput();
            case 3 -> System.out.println(itemRegistry.keySet());
            case 4 ->
            {
                itemTable.displayItemTable(itemRegistry);
                System.out.println("~~~");
                System.out.println("Window with table has appeared! note that you can sort"
                    + " alphabetically by clicking the parameter above");
                System.out.println("~~");
            }
            case 5 -> deleteItemFromItemRegistry();
            case 6 -> filterInItemRegistry();
            case 7 -> sortInItemRegistry();
            case 8 -> printItemRegistryCategory();
            case 9 -> setItemOnSale();
            default ->
            {
                if (userInput != 0)
                {
                    System.out.println("No known command was found with "
                        + "your input please retry");
                }
            }
        }
        return userInput;
    }

    /**
     * Gives the user options for what they wish to do with the itemRegistry.
     *
     */
    public UserInterface()
    {
        retrieveItemRegistry();
        int userInput;
        do
        {
            userInput = -1;
            try
            {
                userInput = userOptionHandler();
            } catch (NumberFormatException e)
            {
                System.err.println("You have typed a non-number or decimal, please retry");
            } catch (Exception e)
            {
                System.err.println("An unexpected exception has occurred");
            }
        } while (userInput != 0);
        save();
        System.out.println("Exiting...");
    }
}

